"""Module containing the Builder class."""
from __future__ import annotations
from os import PathLike

import build123d as _

from .mutation import Mutation
from .utils import ColorLike, Hash
from .shapes import FaceListLike, EdgeListLike, FaceDict, EdgeDict


class Builder:
    """A class used to manipulated Build123d objects that keeps track of each
    performed mutation and manage shape colors."""

    debug_alpha = 0.2
    "The alpha values used for translucent shapes in debug mode."

    default_color: ColorLike = "orange"
    "The default color to be used when a color is passed to a mutation"

    def __init__(self, part: _.Part, color: ColorLike|None=None, debug=False):
        self.object = part
        self.mutations: list[Mutation] = []
        self.debug_faces: dict[Hash, ColorLike] = {}
        self.mutate(self.__class__.__name__, part, color, debug)

    def __getitem__(self, mut_idx: int):
        return self.mutations[mut_idx]

    def __call__(self) -> list[_.Face]:
        faces = self.mutations[-1].faces
        faces_color = self.get_faces_color()

        for face_hash, face in faces.items():
            face.color = faces_color[face_hash] or self.default_color
            face.label = face_hash[:6]

        return list(faces.values())

    def __mul__(self, location: _.Location) -> Builder:
        return Builder(location * self.object)

    def get_face_mutation(self, face: _.Face|Hash) -> Mutation:
        """Retrieve the mutation who created the given face."""

        _hash = Mutation.hash_shape(face) if isinstance(face, _.Face) else face
        for mutation in self.mutations:
            if _hash in mutation.faces:
                return mutation
        raise ValueError

    def get_edge_mutation(self, edge: _.Edge|Hash) -> Mutation:
        """Retrieve the mutation who created the given edge."""

        _hash = Mutation.hash_shape(edge) if isinstance(edge, _.Edge) else edge
        for mutation in self.mutations:
            if _hash in mutation.edges:
                return mutation
        raise ValueError

    def get_faces_color(self) -> dict[Hash, ColorLike|None]:
        """Return a dictionnary containing the color of each face of the current
        object."""

        faces_color: dict[Hash, ColorLike|None] = {}

        for mut in self.mutations:

            for face_hash in mut.faces_added:
                color = mut.color
                if mut.faces_alias:
                    old_hash = mut.faces_alias[face_hash]
                    color = faces_color[old_hash]

                faces_color[face_hash] = color

            rm_colors = {faces_color[rm_hash] for rm_hash in mut.faces_removed}

            if len(rm_colors) == 1:
                rm_color = rm_colors.pop() if len(rm_colors) == 1 else None
                for face_hash in mut.faces_altered:
                    faces_color[face_hash] = rm_color
            else:
                for al_hash, al_face in mut.faces_altered.items():
                    for rm_hash, rm_face in mut.faces_removed.items():
                        if Mutation.is_altered_faces(al_face, rm_face):
                            faces_color[al_hash] = faces_color[rm_hash]

        if self.debug_faces:
            for face_hash, color in faces_color.items():
                if face_hash in self.debug_faces:
                    faces_color[face_hash] = self.debug_faces[face_hash]
                else:
                    r, v, b = self.cast_color(color).to_tuple()[:3]
                    faces_color[face_hash] = _.Color(r, v, b, self.debug_alpha)

        return faces_color

    def get_mutation(self, mutation_id: str) -> Mutation|None:
        """Return the mutation identified by the given id."""

        for mutation in self.mutations:
            if mutation.id == mutation_id:
                return mutation
        return None

    @classmethod
    def _cast_faces(cls, faces: FaceListLike, fake_hashes=False) -> FaceDict:
        """Cast the given faces to a FaceDict. If hashes are not used, set
        `fake_hashes` to True for better performance."""

        if isinstance(faces, FaceDict):
            return faces

        if isinstance(faces, _.Face):
            face_hash = 'f' if fake_hashes else Mutation.hash_shape(faces)
            return FaceDict({face_hash: faces})

        faces_dict = FaceDict({})
        for idx, face in enumerate(faces):
            face_hash = str(idx) if fake_hashes else Mutation.hash_shape(face)
            faces_dict[face_hash] = face
        return faces_dict

    @classmethod
    def _cast_edges(cls, edges: EdgeListLike, fake_hashes=False) -> EdgeDict:
        """Cast the given edges an EdgeDict. If hashes are not used, set
        `fake_hashes` to True for better performance."""

        if isinstance(edges, EdgeDict):
            return edges

        if isinstance(edges, _.Edge):
            edge_hash = 'f' if fake_hashes else Mutation.hash_shape(edges)
            return EdgeDict({edge_hash: edges})

        edges_dict = EdgeDict({})
        for idx, edge in enumerate(edges):
            edge_hash = str(idx) if fake_hashes else Mutation.hash_shape(edge)
            edges_dict[edge_hash] = edge
        return edges_dict

    @classmethod
    def _cast_part(cls, part: Builder|_.Part) -> _.Part:
        """Cast an EdgeListLike to a Edge iterable."""
        return part if isinstance(part, _.Part) else part.object

    @classmethod
    def cast_color(cls, color: ColorLike) -> _.Color:
        """Cast a ColorLike to a Color"""
        return color if isinstance(color, _.Color) else _.Color(color)

    @classmethod
    def _part_color(cls, part: Builder|_.Part) -> ColorLike|None:
        """Retrieve the color of the current object."""

        if isinstance(part, Builder) and len(part.mutations) == 1:
            return part.mutations[-1].color
        return None

    def mutate(
            self,
            name: str,
            obj: _.Part,
            color: ColorLike|None,
            debug: bool,
            faces_alias: dict[Hash, Hash]|None=None
        ) -> Mutation:
        """Base mutation: mutate the current object to the given one by applying
        a mutation with the given name, color and debug mode."""

        self.object = obj

        mutation = Mutation(
            obj,
            self.mutations[-1] if self.mutations else None,
            name,
            len(self.mutations),
            color,
            faces_alias
        )

        if debug:
            for face_hash in mutation.faces_added:
                self.debug_faces[face_hash] = color

        self.mutations.append(mutation)
        return mutation

    def move(self, location: _.Location, color: ColorLike|None=None, debug=False) -> Mutation:
        """Mutation: move the object to the given location, keeping the colors.
        with the given color and debug mode.
        If not color is defined, keep the previous ones for each face."""

        obj = location * self.object
        faces_alias: dict[Hash, Hash] = {}

        for face in self.object.faces():
            old_hash = Mutation.hash_shape(face)
            new_hash = Mutation.hash_shape(location * face)
            faces_alias[new_hash] = old_hash

        return self.mutate('move', obj, color, debug, faces_alias)

    def add(
            self,
            part: Builder|_.Part,
            color: ColorLike|None=None,
            debug=False
        ) -> Mutation:
        """Mutation: fuse the given part to the current object.
        with the given color and debug mode."""

        obj = self.object + self._cast_part(part)
        return self.mutate('add', obj, color or self._part_color(part), debug)

    def sub(
            self,
            part: Builder|_.Part,
            color: ColorLike|None=None,
            debug=False
        ) -> Mutation:
        """Mutation: substract the given part from the current object,
        with the given color and debug mode."""

        obj = self.object - self._cast_part(part)
        return self.mutate('sub', obj, color or self._part_color(part), debug)

    def fillet(
            self,
            edges: EdgeListLike,
            radius: float,
            color: ColorLike|None=None,
            debug=False
        ) -> Mutation:
        """Mutation: apply a fillet of the given radius to the given edges of
        the current object, with the given color and debug mode."""

        edges = self._cast_edges(edges, fake_hashes=True)()
        obj = self.object.fillet(radius, edges)
        return self.mutate('fillet', obj, color, debug)

    def chamfer(
            self,
            edges: EdgeListLike,
            length: float,
            length2: float|None=None,
            face: _.Face|None=None,
            color: ColorLike|None=None,
            debug=False
        ) -> Mutation:
        """Mutation: apply a chamfer of the given length to the given edges of
        the current object, with the given color and debug mode."""

        edges = self._cast_edges(edges, fake_hashes=True)()
        obj = self.object.chamfer(length, length2, edges, face) # type: ignore
        return self.mutate('chamfer', obj, color, debug)

    def debug(self, faces: FaceListLike, color: ColorLike="red"):
        """Set a face for debugging, so it will appear in the given color while
        the rest of the object will be translucent."""

        for face_hash in self._cast_faces(faces):
            self.debug_faces[face_hash] = color

    def export(
            self,
            exporter: _.Export2D,
            file_path: PathLike|bytes|str,
            include_part=True
        ):
        """Export the current object using the given exporter in the given file
        path. If `include_part` is false, do not include the object."""

        if include_part:
            exporter.add_shape(self.object) # type: ignore
        exporter.write(file_path) # type: ignore

    def export_stl(
            self,
            file_path: PathLike|bytes|str,
            tolerance: float = 0.001,
            angular_tolerance: float = 0.1,
            ascii_format: bool = False
        ):
        """Export the current object in STL format to the given file path,
        with the given tolerance, angular tolerance and ascii format mode."""
        _.export_stl(self.object, file_path, tolerance, angular_tolerance, ascii_format)



"""A module used to store utility stuff used in various other modules."""
from typing import TypeAlias

import build123d as _

Hash: TypeAlias = str

ColorLike: TypeAlias = (
    _.Color | # build123d color
    _.Quantity_ColorRGBA | # OCP color
    str | # name, ex: "red"
    tuple[str, int] | # name + alpha, ex: ("red", 0.5)
    tuple[float, float, float] | # rvb, ex: (1, 0, 0)
    tuple[float, float, float, int] | # rvb + alpha, ex: (1, 0, 0, 0.5)
    int | # hexa, ex: 0xff0000
    tuple[int, int] # hexa + alpha, ex: (0xff0000, 0x80)
)



"""A module used to store shapes-related stuff"""
from enum import Enum
from typing import TypeAlias, Iterable

import build123d as _

from .utils import Hash


class ShapeState(Enum):
    """The possible states of a shape for a mutation."""

    ADDED = 1
    ALTERED = 2
    UNTOUCHED = 3
    REMOVED = 4


class EdgeDict(dict):
    """A custom dictionnary used to store edges by their hash.
    If the dict is called ie. `my_edges()`, a list is returned."""

    def __init__(self, edges_dict: dict[Hash, _.Edge]):
        super().__init__(edges_dict)

    def __setitem__(self, edge_hash: Hash, edge: _.Edge) -> None:
        super().__setitem__(edge_hash, edge)

    def __getitem__(self, edge_hash: Hash) -> _.Edge:
        return super().__getitem__(edge_hash)

    def __call__(self) -> list[_.Edge]:
        return list(self.values())


class FaceDict(dict):
    """A custom dictionnary used to store faces by their hash.
    If the dict is called ie. `my_edges()`, a list is returned."""

    def __init__(self, faces_dict: dict[Hash, _.Face]):
        super().__init__(faces_dict)

    def __setitem__(self, face_hash: Hash, face: _.Face) -> None:
        super().__setitem__(face_hash, face)

    def __getitem__(self, face_hash: Hash) -> _.Face:
        return super().__getitem__(face_hash)

    def __call__(self) -> list[_.Face]:
        return list(self.values())


FaceListLike: TypeAlias = FaceDict | Iterable[_.Face] | _.Face
EdgeListLike: TypeAlias = EdgeDict | Iterable[_.Edge] | _.Edge



"""Module containing the Mutation class."""
from __future__ import annotations
from hashlib import md5

import build123d as _

from .utils import ColorLike, Hash
from .shapes import ShapeState, FaceDict, EdgeDict


class Mutation:
    """Class managing the mutation applied when mutating the object."""

    def __init__(
        self,
        obj: _.Part,
        previous: Mutation|None,
        name: str,
        index: int,
        color: ColorLike|None,
        faces_alias: dict[Hash, Hash]|None
    ) -> None:
        self.previous = previous
        self.name = name
        self.index = index
        self.color = color
        self.faces_alias = faces_alias or {}

        self.id = f"{ name }-{ index }"

        self.faces = FaceDict({self.hash_shape(f): f for f in obj.faces()})
        self.faces_state = self.get_faces_state()
        self.faces_added = self.filter_faces(ShapeState.ADDED)
        self.faces_altered = self.filter_faces(ShapeState.ALTERED)
        self.faces_untouched = self.filter_faces(ShapeState.UNTOUCHED)
        self.faces_removed = self.filter_faces(ShapeState.REMOVED)

        self.edges = EdgeDict({self.hash_shape(e): e for e in obj.edges()})
        self.edges_state = self.get_edges_state()
        self.edges_added = self.filter_edges(ShapeState.ADDED)
        self.edges_altered = self.filter_edges(ShapeState.ALTERED)
        self.edges_untouched = self.filter_edges(ShapeState.UNTOUCHED)
        self.edges_removed = self.filter_edges(ShapeState.REMOVED)

        self.vertices = {self.hash_shape(v): v for v in obj.vertices()}

    def __repr__(self):
        return self.id

    @classmethod
    def hash_shape(cls, shape: _.Shape) -> Hash:
        """Return a reproducible hash.
        OCP 7.2 might produce better hashes that could make this unnecessary."""

        def to_int(number: float) -> int:
            return int(number * 1000)

        def serialize_vertex(vertex: _.Vertex) -> tuple:
            return tuple(to_int(v) for v in vertex.to_tuple())

        def serialize_edge(edge: _.Edge) -> tuple:
            vertices = tuple(serialize_vertex(v) for v in edge.vertices())
            is_circle = edge.geom_type == _.GeomType.CIRCLE
            radius = to_int(edge.radius) if is_circle else 0
            return (edge.geom_type, vertices, radius)

        def serialize_face(face: _.Face) -> tuple:
            return tuple(serialize_edge(edge) for edge in face.edges())

        def serialize_part(part: _.Part) -> tuple:
            return tuple(serialize_face(face) for face in part.faces())

        if isinstance(shape, _.Vertex):
            serialized = serialize_vertex(shape)
        elif isinstance(shape, _.Edge):
            serialized = serialize_edge(shape)
        elif isinstance(shape, _.Face):
            serialized = serialize_face(shape)
        elif isinstance(shape, _.Part):
            serialized = serialize_part(shape)
        else:
            raise TypeError

        return md5(str(serialized).encode()).hexdigest()

    def filter_faces(self, state: ShapeState) -> FaceDict:
        """Return the faces of the current object that match the given state."""

        faces = (
            self.previous.faces
            if self.previous and state == ShapeState.REMOVED
            else self.faces
        )
        faces = {h: faces[h] for h, s in self.faces_state.items() if s == state}
        return FaceDict(faces)

    def filter_edges(self, state: ShapeState) -> EdgeDict:
        """Return the edges of the current object that match the given state."""

        edges = (
            self.previous.edges
            if self.previous and state == ShapeState.REMOVED
            else self.edges
        )
        edges = {h: edges[h] for h, e in self.edges_state.items() if e == state}
        return EdgeDict(edges)

    @classmethod
    def is_altered_faces(cls, this_face: _.Face, that_face: _.Face):
        """Check if the two given faces were altered or not, by comparing the
        edges of each face: if a similar edge is found, they are altered."""

        for this_edge in this_face.edges():
            this_hash = cls.hash_shape(this_edge)
            for that_edge in that_face.edges():
                if this_hash == cls.hash_shape(that_edge):
                    return True


        if (
            this_face.geom_type == that_face.geom_type
            and this_face.location == that_face.location
            and this_face.center_location == that_face.center_location
        ):
            return True

        return False

    def is_altered_face(self, face: _.Face):
        """Check if the given face were altered, by comparing the edges of the
        face: if a similar edge is found in the object, it is altered."""

        if not self.previous:
            return True

        for edge in face.edges():
            if self.hash_shape(edge) in self.previous.edges:
                return True

        for that_face in self.previous.faces():
            if (face.geom_type == that_face.geom_type
                and face.location == that_face.location
                and face.center_location == that_face.center_location
            ):
                return True

        return False

    def get_faces_state(self) -> dict[Hash, ShapeState]:
        """Return a dictionnary holding the state of each face of the object."""

        def get_state(face_hash: Hash, face: _.Face) -> ShapeState:
            if not self.previous:
                return ShapeState.ADDED

            if face_hash in self.previous.faces:
                return ShapeState.UNTOUCHED

            if self.is_altered_face(face):
                return ShapeState.ALTERED

            return ShapeState.ADDED

        faces = {fh: get_state(fh, face) for fh, face in self.faces.items()}

        if self.previous:
            for face_hash in self.previous.faces:
                if face_hash not in self.faces:
                    faces[face_hash] = ShapeState.REMOVED

        return faces

    def is_altered_edge(self, edge: _.Edge):
        """Check if the given edge were altered, by comparing the vertices of
        the face: if a similar vertex is found in the object, it is altered."""

        if not self.previous:
            return True

        for vertex in edge.vertices():
            if self.hash_shape(vertex) in self.previous.vertices:
                return True

        return False

    def get_edges_state(self) -> dict[Hash, ShapeState]:
        """Return a dictionnary holding the state of each edge of the object."""

        def get_state(edge_hash: Hash, edge: _.Edge) -> ShapeState:
            if not self.previous:
                return ShapeState.ADDED

            if edge_hash in self.previous.edges:
                return ShapeState.UNTOUCHED

            if self.is_altered_edge(edge):
                return ShapeState.ALTERED

            return ShapeState.ADDED

        edges = {eh: get_state(eh, edge) for eh, edge in self.edges.items()}

        if self.previous:
            for edge_hash in self.previous.edges:
                if edge_hash not in self.edges:
                    edges[edge_hash] = ShapeState.REMOVED

        return edges
